"""
Author: Nicolas Mauchle <mauchle@semafor.ch>
To use this class you have to install:

- gcloud
- googleapiclient

Also you have to configure your google account with
  gcloud init

The google cloud module for femag in short:
1. Create for every calculation a bucket
2. Upload the files for the calculation in the specific bucket
3. Start for every calculation one instance with a femag image
4. Wait until a exit_code file is stored in the bucket
5. Terminate instance
6. Sync back the results files

The connection between the buckets and the calculation are stored in the job dict

"""
import logging
import os
import threading      # Not supported from httplib
import time
import femagtools
import random         # later used in create project
import pdb
from .config import Config

logger = logging.getLogger(__name__)

GOOGLE_MODULE_FOUND = True
try:
    from oauth2client.client import GoogleCredentials
    from googleapiclient import discovery
    from gcloud import storage
    from gcloud import resource_manager
except:
    GOOGLE_MODULE_FOUND = False


class IllegalNumberOfBuckets(Exception):
    def __init__(self):
        Exception.__init__(self)
        logger.error("Max 10 buckets are allowed")

class Engine():

    config_class = Config

    default_config = {
        'ENGINE': 'google',
        'SERVER_LOCATION': 'us-east1-b',  # 'f1-micro'
        'INSTANCE_TYPE': 'n1-standard-1',
        'IMAGE_ID': 'femag-v1',
        'FINISH_TASK_FILENAME': 'exit_code',
        'COMPANY_NAME': 'femag',
    }

    def __init__(self, buckets=None):
        """Initalize the Google cloud service

        Args:
            buckets:     to prevent uploads and use these buckets with data
        """
        if not GOOGLE_MODULE_FOUND:
            logger.error("Could not find one of [google-api-python-client, gcloud, googleapis-common-protos] modules")
            return

        if buckets and len(buckets) > 10:
            raise IllegalNumberOfBuckets()

        self.buckets = buckets
        self.job = None

        self.config = Config(self.default_config)

        # Create a new project
        self.project = self._create_new_project('ancient-ship-136307')

    def _create_data_buckets(self):
        """Every calculation has its own bucket with all data
        All buckets are created before any calculation.
        The name of a bucket is unique uuid number

        .. note::  We can only create 10 buckets in a short time period
        """

        # If we have already some buckets we have to assign
        # them to our jobs
        if self.buckets:
            logger.info("Buckets already exists")
            for idx, bucket in enumerate(self.buckets):
                self.job.tasks[idx].id = bucket['id']
                self.job.tasks[idx].directory = bucket['folder']
            return

        gcs = storage.Client(self.project.project_id)
        # Create new buckets for calculation
        logger.info("Create {} new buckets".format(len(self.job.tasks)))
        for t in self.job.tasks:
            gcs.create_bucket(t.id)

        logger.info("Buckets created")

    def _upload_files_to_buckets(self):
        """For every calculation there are some files to upload in the specific bucket
        To save time the upload is done in a seperate thread.

        .. note::

        Unfortunately the google storage model can not handle threads (SSL version error)
        To avoid that, we have to create a new storage client in every thread.
        """

        # Do not upload files if whe already have the buckets
        if self.buckets:
            logger.info("Files are already uploaded")
            return

        threads = []
        for t in self.job.tasks:
            logger.info("Start upload folder {} from task {}".format(t.file, t.id))
            thread = threading.Thread(target=self._upload, args=(t, ))
            thread.start()
            threads.append(thread)

        # Wait for all finished uploads
        self._wait_for_threads_finished(threads, "Uploading files")

    def _upload(self, task):
        """Upload the file to the google storage

        Args:
            task:  The task which define which folder (task.directory) should be uploaded
                      to which storage (task.id)
        """
        # We can not use the self.gcs cause we use it in a thread and if we use self.gcs an exception
        # is thrown -> Wrong SSL Version.
        gcs = storage.Client(self.project.project_id)

        # task.id is the uuid generated name for the bucket
        bucket = gcs.get_bucket(task.id)
        # Close the tar file
        task.tar_file.close()
        
        blob = storage.Blob(os.path.basename(task.file), bucket)
        with open(task.file, 'rb') as file:
                blob.upload_from_file(file)

        # Other possibility:
        # bucket = gcs.create_bucket('bucket-name')
        # blob = bucket.blob(filename)
        # blob.upload_from_string(uploaded_file.read(), content_type=uploaded_file.content_type )

    def _create_new_project(self, name):
        """At the moment it is not possible to create a new project overt the python api
        So we use always the default project

        Args:
            name (str):     The name of the project

        Return:
            Google Cloud project (:obj:`Google Cloud Project`)
        """
        gcrm = resource_manager.Client()

        # Create new project
        # project_id = "{}-{}".format(self.company, random.randint(1, 10000))
        # self.project = self.gcrm.new_project(project_id, name=name)
#        project.labels = {'tasks': ", ".join([t.id for t in self.job.tasks]), 'status': 'initalize'}
        # self.project.create()  # Create the project

        project = gcrm.fetch_project(name)
        logger.info("Created new project. ID: {}, Name: {}".format(project.project_id, project.name))
        return project

    def _start_instances(self):
        """ Prepare and start all instances for the calculation.

        """

        # Configuration for our machine. Here we put all our predefined configuration
        import json
        config = json.load(open(self.config['SERVER_CONFIG']))

        # If we want to use a debian image file from google:
        # image_response = self.compute.images().getFromFamily(project='debian-cloud',
        #                                                     family=self.image_id).execute()
        # source_disk_image = image_response['selfLink']

        # Use or own image disk with preinstalled femag and dependencies:
        for disk in config['disks']:
            if 'initializeParams' not in disk:
                disk['initializeParams'] = {'sourceImage': "projects/{}/global/images/{}".format(self.project.project_id, self.config['IMAGE_ID'])}

        if 'machineType' not in config:
            config['machineType'] = "zones/{}/machineTypes/{}".format(self.config['SERVER_LOCATION'],
                                                                      self.config['INSTANCE_TYPE'])

        config['metadata']['items'].append({'key': 'startup-script',
                                            'value': open(self.config['STARTUP_SCRIPT'], 'r').read()})
        # Startup instances and the calculation
        operations = []

        # You always have to be logged in. With GoogleCredentials.get_application_default()
        # it looks in the home directory for the credentials and use the default profile
        credentials = GoogleCredentials.get_application_default()

        # Get the compute object
        compute = discovery.build('compute', 'v1', credentials=credentials)
        for t in self.job.tasks:
            # Set same name for the instance as for the bucket.
            # Googlecloud does not allow instance name which starts with a number
            config['name'] = "{}-{}".format(self.config['COMPANY_NAME'].lower(), t.id)

            # Check if we defined the key 'bucket' in items
            # in loop 1 the bucket is not defined -> so we have to add it
            # after that we overwrite the bucket value
            # Every instance does download his bucket
            idx = [i for i, item in enumerate(config['metadata']['items']) if item['key'] == 'bucket']
            if idx:
                config['metadata']['items'][idx[0]] = {'key': 'bucket', 'value': t.id}
            else:
                config['metadata']['items'].append({'key': 'bucket', 'value': t.id})

            # Start the instance with our project_id zone and configuration
            operation = compute.instances().insert(
                project=self.project.project_id,
                zone=self.config['SERVER_LOCATION'],
                body=config).execute()

            operations.append(operation)

        # Wait until all instances are started up
        self._wait_for_operations_finished(operations)

    def _wait_for_threads_finished(self, threads, operation):
        """This generic methods waits until all threads are finished

        Args:
            list threads:    List of threads to check if they are finished
            str operation:   Name of the operation to write a meaningful log message

        """
        # Wait until all threads are not running
        while not all([t.isAlive() is False for t in threads]):
            time.sleep(5)

        # timer.cancel
        logger.info("{} are finished".format(operation))

    def _wait_for_operations_finished(self, operations):
        """This methods waits until all operations (started on google cloud) are finished

        Args:
            operations (:obj:`list`): A list of google cloud operations
        """
        threads = []
        # timeout = 300
        # timer = threading.Timer(timeout, thread.interrupt_main)
        for operation in [o['name'] for o in operations]:
            thread = threading.Thread(target=self._wait_for_operation, args=(operation, ))
            threads.append(thread)
            thread.start()

        # timer.start()
        self._wait_for_threads_finished(threads, "Operations")

    def _wait_for_operation(self, operation):
        """Wait until a operation is finished
        Is called from :py:meth:`_wait_for_operations_finished`

        Args:
            Operation One Google Operation Object
        """

        #  Do not use self.compute: SSL VERSION Exception when called in thread
        credentials = GoogleCredentials.get_application_default()
        compute = discovery.build('compute', 'v1', credentials=credentials)
        logger.info('Waiting for operation to finish...')
        while True:
            result = compute.zoneOperations().get(
                project=self.project.project_id,
                zone=self.config['SERVER_LOCATION'],
                operation=operation).execute()

            # Check if operation is DONE or has any error
            if result['status'] == 'DONE':
                if 'error' in result:
                    raise Exception(result['error'])

                return result

            time.sleep(10)

    def _delete_instances(self, instances):
        """Delete one ore more instances

        Args:
            instances (:obj:`list`):   A list of instances/task to be deleted

        """
        # You always have to be logged in. With GoogleCredentials.get_application_default()
        # it looks in the home directory for the credentials and use the default profile
        credentials = GoogleCredentials.get_application_default()

        # Get the compute object
        compute = discovery.build('compute', 'v1', credentials=credentials)
        operations = []
        for task in instances:
            logger.info("Delete instace: {}".format(task.id))
            # Be sure to make all strings lower cause of the google regex
            # (?:[a-z](?:[-a-z0-9]{0,61}[a-z0-9])?)
            instance_name = "{}-{}".format(self.config['COMPANY_NAME'].lower(), task.id)
            operations.append(compute.instances().delete(
                project=self.project.project_id,
                zone=self.config['SERVER_LOCATION'],
                instance=instance_name).execute())

        self._wait_for_operations_finished(operations)

    def _download_bucket(self, task):
        """Download all files from one bucket bucket name is task.id in a thread

        Args:
            task (:py:class:`CloudTask`):  The task which is assigned to this bucket to get the bucket name
        """
        # !Threads!
        gcs = storage.Client(self.project.project_id)
        bucket = gcs.get_bucket(task.id)

        if not os.path.exists(task.directory):
            os.makedirs(task.directory)

        # Get all files which are stored in our bucket
        files = bucket.list_blobs()
        for file in files:
            dest = "{}/{}".format(task.directory, file.name)
            # blob = storage.Blob(file.name, bucket)
            # Download the file
            with open(dest, 'wb') as file_obj:
                file.download_to_file(file_obj)

            # Untar the file
#            tar = tarfile.TarFile(dest, 'r:gz')
#            tar.extractall(os.path.dirname(dest))

    def _download_results_from_googlecloud(self):
        """Download all files from all instances/tasks to our result folder for femag

        """
        threads = []
        logger.info("Start downloading files")
        for t in self.job.tasks:
            thread = threading.Thread(target=self._download_bucket, args=(t, ))
            threads.append(thread)
            thread.start()

        self._wait_for_threads_finished(threads, "Downloading files")

    def _get_status_code(self, filename='exit_code'):
        """Get the status code from the caluclation
        Status code is written in a file

        Args:
            filename (str): The filename where the exit code is stored
        """
        status_code = []
        for t in self.job.tasks:
            dir = "{}/{}".format(t.directory, filename)
            file = open(dir, 'r')
            status_code.append(file.read())
        return status_code

    def _join(self, filename='exit_code', delete=True):
        """Wait until all calculation are finished:
        This means wait until every bucket has an file with the name of the given filename

        Args:
            filename (str): The filename where the exit code is stored
            delete (bool):  Should the instance be deleted after calculation
        """
        finished_tasks = []
        logger.info("Calculating..")
        gcs = storage.Client(self.project.project_id)
        while len(finished_tasks) < len(self.job.tasks):
            for t in [task for task in self.job.tasks if task not in finished_tasks]:
                bucket = gcs.get_bucket(t.id)
                if bucket.get_blob(filename):
                    finished_tasks.append(t)
                    logger.info("Calculation is finished for instance: {}".format(t.id))
                    if delete:
                        self._delete_instances([t])

            time.sleep(10)

        logger.info("Calculation is finished")

    def _cleanup(self):
        """Remove all buckets which belongs to this calculation round

        """
        threads = []
        for t in self.job.tasks:
            thread = threading.Thread(target=self._delete_bucket, args=(t.id, ))
            threads.append(thread)
            thread.start()

        logger.info("Deleting buckets: ")
        self._wait_for_threads_finished(threads, "Deleting buckets")

    def _delete_bucket(self, bucket_name):
        gcs = storage.Client(self.project.project_id)
        bucket = gcs.get_bucket(bucket_name)
        bucket.delete_blobs(bucket.list_blobs())
        bucket.delete(force=True)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # FEMAG STUFF
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def create_job(self, workdir):
        """Create a FEMAG :py:class:`CloudJob`

        Args:
            workdir (str): The workdir where the calculation files are stored

        Return:
            Cloud job (:py:class:`CloudJob`)
        """
        self.job = femagtools.job.CloudJob(workdir)
        return self.job

    def submit(self):
        """Starts the FEMAG calculation(s) on Google Cloud

        Return:
            length of started tasks (int)
        """
        self._create_data_buckets()
        self._upload_files_to_buckets()
        self._start_instances()
        return len(self.job.tasks)

    def join( self ):
        """Wait until all calculations are finished

        Return:
            list of all calculations status (C = Ok, X = error) (:obj:`list`)
        """
        self._join(filename='exit_code', delete=True)

        self._download_results_from_googlecloud()

        if int(self.config.get('DELETE_BUCKETS', 0)):
            self._cleanup()
        status = self._get_status_code("exit_code")
        for t, r in zip(self.job.tasks, status):
            t.status = 'C' if int(r)==0 else 'X'

        return status


if __name__ == "__main__":
    g_engine = Engine()
    pdb.set_trace()  # Set breakpoint



"""
    femagtools.docker
    ~~~~~~~~~~~~~~~~~

    Running FEMAG on Docker/Cloud


"""
import os
import json
import logging
import threading
import femagtools.femag
import femagtools.job
import time
try:
    from queue import Queue
except ImportError:
    from Queue import Queue  # python 2.7


logger = logging.getLogger(__name__)


def publish_receive(message):
    """handle messages from femag publisher"""
    topic, content = message  # "femag_log" + text
    # topics: femag_log, progress, file_modified,
    #   model_image, calc_image, field_image, babs_image, demag_image, color_scale
    if topic == 'femag_log' or topic == 'progress':
        logger.info("%s: %s", topic, content.strip())
    else:
        logger.info('%s: len %d', topic, len(content.strip()))


class AsyncFemag(threading.Thread):
    def __init__(self, queue, port, host):
        threading.Thread.__init__(self)
        self.queue = queue
        self.container = femagtools.femag.ZmqFemag(
            port, host)

    def _do_task(self, task):
        r = self.container.cleanup(timeout=10000)
        status = json.loads(r[0])
        if status['status'] != 'ok':
            return [status]
        for f in task.transfer_files:
            if f != task.fsl_file:
                r = self.container.upload(
                    os.path.join(task.directory, f))
                status = json.loads(r[0])
                if status['status'] != 'ok':
                    return [status]
            fslfile = os.path.join(task.directory, task.fsl_file)
        logger.info('Docker task %s %s',
                    task.id, task.fsl_file)
        fslcmds = []
        with open(fslfile) as f:
            fslcmds = f.readlines()
        ret = self.container.send_fsl(fslcmds +
                                      ['save_model(close)'])
        # TODO: add publish_receive
        return [json.loads(s) for s in ret]
        
    def run(self):
        """execute femag fsl task in task directory"""
        while True:
            task = self.queue.get()
            if task is None:
                break
            while True:
                r = self._do_task(task)
                if r[0]['status'] != 'resend':
                    break
                time.sleep(1)

            logger.debug("Finished %s", r)
            try:
                if r[0]['status'] == 'ok':
                    task.status = 'C'
                    bchfile = r[0]['result_file'][0]
                    status, content = self.container.getfile(bchfile)
                    logging.info("get results %s: status %s len %d",
                                 task.id, status, len(content))
                    with open(os.path.join(task.directory,
                                           bchfile), 'wb') as f:
                        f.write(content)
                else:
                    task.status = 'X'
                    logger.warn("%s: %s", task.id, r[0]['message'])
            except (KeyError, IndexError):
                task.status = 'X'

            logger.info("Task %s end status %s",
                        task.id, task.status)
            ret = self.container.release()
            self.queue.task_done()
        self.container.close()


class Engine(object):

    """The Docker Engine

       execute Femag-Simulations with docker

       Args:
         dispatcher (str): hostname of dispatcher
         port (int): port number of dispatcher
         num_threads: number of threads to send requests
    """
    def __init__(self, dispatcher='127.0.0.1', port=5000,
                 num_threads=5):
        self.port = port
        self.dispatcher = dispatcher
        self.num_threads = num_threads
        self.async_femags = None
        
    def create_job(self, workdir):
        """Create a FEMAG :py:class:`CloudJob`

        Args:
            workdir (str): The workdir where the calculation files are stored

        Return:
            job (:class:`Job`)
        """
        self.job = femagtools.job.Job(workdir)
        return self.job

    def submit(self):
        """Starts the FEMAG calculation(s) as Docker containers

        Return:
            number of started tasks (int)
        """
        self.queue = Queue()
        for task in self.job.tasks:
            self.queue.put(task)
            
        logger.info("Request %d workers on %s",
                    self.num_threads, self.dispatcher )
        self.async_femags = [AsyncFemag(self.queue,
                                        self.port, self.dispatcher)
                             for i in range(self.num_threads)]

        for async_femag in self.async_femags:
            async_femag.start()

        return len(self.job.tasks)

    def join(self):
        """Wait until all calculations are finished

        Return:
            list of all calculations status (C = Ok, X = error) (:obj:`list`)
        """
        # block until all tasks are done
        self.queue.join()

        # stop workers
        for _ in self.async_femags:
            self.queue.put(None)

        # wait for all workers
        for async_femag in self.async_femags:
            async_femag.join()

        return [t.status for t in self.job.tasks]



# -*- coding: utf-8 -*-
"""
    femagtools.femag
    ~~~~~~~~~~~~~~~~

    Running FEMAG


"""
import subprocess
import os
import glob
import logging
try:
    import zmq
except ImportError:
    pass
import sys
import json
import io
import femagtools.mcv
import femagtools.airgap as ag
import femagtools.fsl
import femagtools.ntib as ntib
import femagtools.config as cfg
import time
import platform
import re
from threading import Thread

logger = logging.getLogger(__name__)


BCHEXT = '.BATCH' if sys.platform.startswith('linux') else '.BCH'  # win32

def get_shortCircuit_parameters(bch, nload):
    try:
        if nload < 0: nload=0
        if nload > 2: nload=2
        if nload > 0:
            dqld = bch.dqPar['ld']
            dqlq = bch.dqPar['lq']
            dqpsim = bch.dqPar['psim']
            if len(dqld) <= nload or len(dqlq) <= nload or len (dqpsim) <= nload:
                ld = dqld[-1]/bch.armatureLength
                lq = dqlq[-1]/bch.armatureLength
                psim = dqpsim[-1]/bch.armatureLength
            else:
                ld = dqld[nload-1]/bch.armatureLength
                lq = dqlq[nload-1]/bch.armatureLength
                psim = dqpsim[nload-1]/bch.armatureLength
        else:
            ld = bch.machine['ld']/bch.armatureLength
            lq = bch.machine['lq']/bch.armatureLength
            psim = bch.machine['psim']/bch.armatureLength
        return dict(
                r1 = bch.machine['r1'],
                ld = ld,
                lq = lq,
                psim = psim,
                num_pol_pair = bch.machine['p'],
                fc_radius = bch.machine['fc_radius'],
                lfe = bch.armatureLength/1e3,
                pocfilename = bch.machine['pocfile'],
                num_par_wdgs = bch.machine['num_par_wdgs'],
                calculationMode = 'shortcircuit')
    except (KeyError, AttributeError, IndexError):
        raise FemagError("missing pm/Rel-Sim results")
    
class FemagError(Exception):
    pass


def subscribe_dev_null(message):
    logger.debug("DevNull: %s", message)
    return


class BaseFemag(object):
    def __init__(self, workdir, cmd, magnetizingCurves, magnets):
        self.workdir = workdir
        if cmd:
            self.cmd = cmd
        else:
            self.cmd = cfg.get_femag()

        if magnetizingCurves:
            if isinstance(magnetizingCurves,
                          femagtools.mcv.MagnetizingCurve):
                self.magnetizingCurves = magnetizingCurves
            else:
                self.magnetizingCurves = femagtools.mcv.MagnetizingCurve(
                    magnetizingCurves)
        else:
            self.magnetizingCurves = []
        if magnets:
            if isinstance(magnets, femagtools.Magnet):
                self.magnets = magnets
            else:
                self.magnets = femagtools.Magnet(magnets)
        else:
            self.magnets = []

    def copy_magnetizing_curves(self, model, dir=None):
        """extract mc names from model and write files into workdir or dir if given

        Return:
            list of extracted mc names (:obj:`list`)
        """
        dest = dir if dir else self.workdir
        return [self.magnetizingCurves.writefile(m[0], dest, fillfac=m[1])
                for m in model.set_magcurves(
                        self.magnetizingCurves, self.magnets)]

    def create_fsl(self, pmMachine, simulation):
        """create list of fsl commands"""
        self.model = femagtools.MachineModel(pmMachine)
        self.modelname = self.model.name
        self.copy_magnetizing_curves(self.model)

        builder = femagtools.fsl.Builder()
        if simulation:
            return builder.create(self.model, simulation, self.magnets)
        return builder.create_model(self.model, self.magnets) + ['save_model("cont")']
        

    def get_log_value(self, pattern, modelname='FEMAG-FSL.log'):
        result = []
        pat = re.compile(pattern)
        with open(os.path.join(self.workdir, 'FEMAG-FSL.log')) as f:
            for line in f:
                m = pat.search(line)
                if m:
                    result.append(float(line.split(':')[-1].split()[0]))
        return result

    def get_bch_file(self, modelname, offset=0):
        """return latest bch file (if any)"""
        bchfile_list = sorted(glob.glob(os.path.join(
            self.workdir, modelname+'_[0-9][0-9][0-9]'+BCHEXT)))
        if(bchfile_list):
            return bchfile_list[-1-offset]
        return ''
    
    def read_bch(self, modelname=None, offset=0):
        "read most recent BCH/BATCH file and return result"
        # read latest bch file if any
        if not modelname:
            modelname = self._get_modelname_from_log()

        result = femagtools.bch.Reader()
        bchfile = self.get_bch_file(modelname, offset)
        if bchfile:
            logger.info("Read BCH {}".format(bchfile))
            with io.open(bchfile, encoding='latin1',
                         errors='ignore') as f:
                result.read(f)
        return result

    def read_isa(self, modelname=None):
        "read most recent I7/ISA7 file and return result"
        if not modelname:
            modelname = self._get_modelname_from_log()
        import femagtools.isa7
        return femagtools.isa7.read(os.path.join(self.workdir, modelname))

    def read_los(self, modelname=None):
        "read most recent LOS file and return result"
        # read latest los file if any
        if not modelname:
            modelname = self._get_modelname_from_log()

        losfile_list = sorted(glob.glob(os.path.join(
            self.workdir, modelname+'_[0-9][0-9][0-9].LOS')))
        if len(losfile_list) > 0:
            return ntib.read_los(losfile_list[-1])

        return dict()

    def read_airgap_induction(self, modelname='', offset=0):
        """read airgap induction"""
        # we need to figure out the number of poles in model
        bch = self.read_bch(modelname, offset)
        return ag.read(os.path.join(self.workdir, 'bag.dat'),
                       bch.machine['p_sim'])

    def _get_modelname_from_log(self):
        """
        Read the modelname from the Femag Log file
        """
        with open(os.path.join(self.workdir, 'FEMAG-FSL.log')) as f:
            for l in f:
                if l.startswith('New model') or l.startswith('Load model'):
                    model = l.split('"')[1]
                    break
        return model


class Femag(BaseFemag):
    """Invoke and control execution of FEMAG

    Args:
        workdir: name of working directory
        cmd: name of femag program (default wfemag64 on windows, xfemag64 on linux)
        magnetizingCurves: collection of lamination material curves
        magnets: collection of magnet material
    """
    def __init__(self, workdir, cmd=None,
                 magnetizingCurves=None, magnets=None):
        super(self.__class__, self).__init__(workdir, cmd,
                                             magnetizingCurves, magnets)

    def run(self, filename, options=['-b'], fsl_args=[]):
        """invoke FEMAG in current workdir

        Args:
            filename: name of file to execute
            options: list of FEMAG options
            fsl_args: list of FSL argument options

        Raises:
            FemagError
        """
        if self.cmd.startswith('wfemag') and \
           '-b' in options and \
           '-m' not in options:
            options.insert(0, '-m')
        args = [self.cmd] + options + [filename] + fsl_args

        basename, ext = os.path.splitext(os.path.basename(filename))
        outname = os.path.join(self.workdir, basename+'.out')
        errname = os.path.join(self.workdir, basename+'.err')
        with open(outname, 'w') as out, open(errname, 'w') as err:
            logger.info('invoking %s', ' '.join(args))
            proc = subprocess.Popen(
                args,
                stdout=out, stderr=err, cwd=self.workdir)

        proc.wait()
        errs = []
        # print femag output
        with io.open(outname, encoding='latin1', errors='ignore') as outfile:
            for l in outfile:
                if l.find('ERROR') > -1:
                    errs.append(l.strip())

        rc = proc.returncode
        logger.info("%s exited with returncode %d (num errs=%d)",
                    self.cmd, rc, len(errs))
        if rc != 0 or errs:
            with io.open(errname, encoding='latin1',
                         errors='ignore') as errfile:
                for l in errfile:
                    errs.append(l.strip())
            errs.insert(0, 'Exit code {}'.format(rc))
            raise FemagError(errs)
        
    def cleanup(self):
        "removes all created files in workdir"
        if not os.path.exists(self.workdir):
            return
        cleanfiles = ('*.B*CH', '*.I*7-*', '*.A*7-*', '*.nc-*',
                      '*.dat', '*.LOS', '*.svg', '*.png', '*.hxy')
        # '*.TMC','*.TMO', '*.PROT'):
        for p in cleanfiles:
            for f in glob.glob(os.path.join(self.workdir, p)):
                os.remove(f)

    def __call__(self, pmMachine, simulation={},
                 options=['-b'], fsl_args=[]):
        """setup fsl file, run calculation and return 
        BCH or LOS results if any."""
        fslfile = 'femag.fsl'
        with open(os.path.join(self.workdir, fslfile), 'w') as f:
            f.write('\n'.join(self.create_fsl(pmMachine,
                                              simulation)))
        if simulation:
            if 'poc' in simulation:
                 with open(os.path.join(self.workdir,
                                        simulation['pocfilename']), 'w') as f:
                    f.write('\n'.join(simulation['poc'].content()))
            if simulation['calculationMode'] == "pm_sym_loss":
                with open(os.path.join(self.workdir,
                                       self.modelname+'.ntib'), 'w') as f:
                    f.write('\n'.join(ntib.create(
                        simulation['speed'],
                        simulation['current'],
                        simulation['angl_i_up'])))
                # TODO: add r1, m
 
        self.run(fslfile, options, fsl_args)
        if simulation:
            if simulation['calculationMode'] == "pm_sym_loss":
                return self.read_los(self.modelname)

            bch = self.read_bch(self.modelname)
            if simulation['calculationMode'] == 'pm_sym_fast':
                if simulation.get('shortCircuit', False):
                    logger.info("short circuit simulation")
                    simulation.update(
                        get_shortCircuit_parameters(bch,
                                                    simulation.get('initial',2)))
                    builder = femagtools.fsl.Builder()
                    fslcmds = (builder.open_model(self.model) +
                               builder.create_shortcircuit(simulation))
                    with open(os.path.join(self.workdir, fslfile), 'w') as f:
                        f.write('\n'.join(fslcmds))
                    self.run(fslfile, options)
                    bchfile = self.get_bch_file(self.modelname)
                    if bchfile:
                        bchsc = femagtools.bch.Reader()
                        logger.info("Read BCH {}".format(bchfile))
                        with io.open(bchfile, encoding='latin1',
                                     errors='ignore') as f:
                            bchsc.read(f)
                    bch.scData = bchsc.scData
                    for w in bch.flux:
                        try:
                            bch.flux[w] += bchsc.flux[w]
                            bch.flux_fft[w] += bchsc.flux_fft[w]
                        except (KeyError, IndexError):
                            logging.debug("No additional flux data in sc simulation")
                            break

                    bch.torque += bchsc.torque
                    bch.demag += bchsc.demag
            return bch
        return dict(status='ok', message=self.modelname)


class ZmqFemag(BaseFemag):
    """Invoke and control execution of FEMAG with ZeroMQ

    Args:
        port: port number of req socket
        host: hostname (ip addr)
        workdir: name of working directory
        cmd: name of femag program
    """
    def __init__(self, port, host='localhost', workdir='', cmd=None,
                 magnetizingCurves=None, magnets=None):
        super(self.__class__, self).__init__(workdir, cmd,
                                             magnetizingCurves, magnets)
        self.host = host
        self.port = port
        self.pubhost = ''

        self.request_socket = self.__req_socket()
        self.subscriber_socket = None
        self.proc = None
        self.reader = None

    def close(self):
        if self.reader:
            self.reader.continue_loop = False
        if self.proc:
            self.quit()
        if self.request_socket:
            self.request_socket.close()
            self.request_socket = None
            self.request_socket = self.__req_socket()
        if self.subscriber_socket:
            self.subscriber_socket.close()
            self.subscriber_socket = None

    def __del__(self):
        self.close()
        logger.debug("Destructor ZmqFemag")

    def __req_socket(self):
        """returns a new request client"""
        context = zmq.Context.instance()
        self.request_socket = context.socket(zmq.REQ)
        self.request_socket.connect('tcp://{0}:{1}'.format(
            self.host, self.port))
        return self.request_socket

    def __sub_socket(self):
        """returns a subscriber client"""
        if self.subscriber_socket:
            return self.subscriber_socket
        context = zmq.Context.instance()
        if not self.pubhost:
            self.pubhost = '127.0.0.1'
        self.subscriber_socket = context.socket(zmq.SUB)
        self.subscriber_socket.connect(
            'tcp://{0}:{1}'.format(
                self.pubhost, self.port+1))
        self.subscriber_socket.setsockopt(zmq.SUBSCRIBE, b'')
        self.subscriber_socket.RCVTIMEO = 900  # in milliseconds
        logger.info("Subscriber %s", self.pubhost)
        return self.subscriber_socket

    def __is_process_running(self, procId):
        try:
            import psutil
            return psutil.pid_exists(procId)
        except ModuleNotFoundError:
            pass
        # long version, self made
        try:
            if procId > 0:
                if platform.system() == "Windows":
                    #                   if procId in psutil.get_pid_list():
                    proc = subprocess.Popen(["tasklist"],
                                            stdout=subprocess.PIPE)
                    for l in proc.stdout:
                        ls = l.split()
                        try:
                            if str(procId) == ls[1]:
                                return True
                        except IndexError:
                            continue
                else:
                    if not os.kill(procId, 0):
                        return True
        except OSError as e:
            # No such process
            logger.info("OSError: %s", e)
            return False
        except Exception as e:
            # we cannot check processId
            logger.warn("process check error %s", e)
            return True
        return False

    def __is_running(self):
        """check if FEMAG is running in ZMQ mode

        Return:
            True if FEMAG is running, False otherwise

        """
        if not self.request_socket:
            return False

        # call info() and check result
        try:
            ret = [json.loads(s, strict=False) for s in self.info()]
            return ret[0].get('status') == 'ok'
        except Exception:
            pass
        return False

    def subscribe(self, pubhost):
        self.pubhost = pubhost
        
       #if not self.pubhost:
        #    if self.host != 'localhost':
        #        inforesp = self.info()
        #        self.pubhost = json.loads(inforesp[1])['addr']
        #        logger.info("Connected with %s", self.pubhost)
        #    else:
        #        self.pubhost = '127.0.0.1'

         
    def send_request(self, msg, pub_consumer=None, timeout=None):
        try:
            # Start the reader thread to get information
            self.restartStreamReader(pub_consumer)
            if timeout:
                self.request_socket.setsockopt(zmq.RCVTIMEO, timeout)
                self.request_socket.setsockopt(zmq.LINGER, 0)
            else:
                self.request_socket.setsockopt(zmq.RCVTIMEO, -1)  # default value
                self.request_socket.setsockopt(zmq.LINGER, 30000)  # default value

            while True:
                try:
                    for m in msg[:-1]:
                        self.request_socket.send_string(m, flags=zmq.SNDMORE)
                    if isinstance(msg[-1], list):
                        self.request_socket.send_string('\n'.join(msg[-1]))
                    else:
                        self.request_socket.send_string(msg[-1])
                    return self.request_socket.recv_multipart()
                except zmq.error.Again:
                    pass
                
        except Exception as e:
            #logger.exception("send_request")
            logger.info("send_request: %s Message %s", str(e), msg)
            if timeout:  # only first call raises zmq.error.Again
                logger.info("Femag is not running")
                return [b'{"status":"error", "message":"Femag is not running"}']
            return [b'{"status":"error", "message":"' + str(e).encode() + b'"}']

    def send_fsl(self, fsl, pub_consumer=None, timeout=None):
        """sends FSL commands in ZMQ mode and blocks until commands are processed

        Args:
            fsl: string (or list) of FSL commands
            timeout: The timeout (in milliseconds) to wait for a response
            pub_consumer: callable object to be invoked with publish message 
                      tuple (topic, content)

        Return:
            status
        """
        header = 'FSL'
        logger.debug("Send fsl with fsl: {}, pub_consumer: {}, header: {}".format(
            fsl, pub_consumer, header))

        try:
            # Start the reader thread to get information about the next calculation
            self.restartStreamReader(pub_consumer)

            self.request_socket.send_string(header, flags=zmq.SNDMORE)
            logger.debug("Sent header")
            if isinstance(fsl, list):
                self.request_socket.send_string('\n'.join(fsl))
            else:
                self.request_socket.send_string(fsl)
            logger.debug("Sent fsl wait for response")

            if timeout:
                self.request_socket.setsockopt(zmq.RCVTIMEO, timeout)
                self.request_socket.setsockopt(zmq.LINGER, 0)
            else:
                self.request_socket.setsockopt(zmq.RCVTIMEO, -1)  # default value
                self.request_socket.setsockopt(zmq.LINGER, 30000)  # default value
            import datetime
            startTime = datetime.datetime.now() if timeout else None
            while True:
                try:
                    response = self.request_socket.recv_multipart()
                    time.sleep(.5)  # Be sure all messages are arrived over zmq
                    if pub_consumer and self.reader:
                        self.reader.continue_loop = False
                    return [s.decode('latin1') for s in response]
                except zmq.error.Again as e:
                    logger.info("Again [%s], timeout: %d", str(e), timeout)
                    if startTime:
                        diffTime = datetime.datetime.now() - startTime
                        logger.debug("Diff msec[%d]", diffTime.microseconds)
                        if diffTime.microseconds < timeout:
                            continue
                        else:
                            logger.info("ALERT not running close socket")
                            self.close()
                            return ['{"status":"error", "message":"Femag is not running"}', '{}']

                    continue
        except Exception as e:
            logger.exception("send_fsl")
            logger.error("send_fsl: %s", str(e))
            if timeout:  # only first call raises zmq.error.Again
                return ['{"status":"error", "message":"Femag is not running"}', '{}']
            msg = str(e)
            return ['{"status":"error", "message":"'+msg+'"}', '{}']

    def run(self, options=['-b'], restart=False, procId=None):  # noqa: C901
        """invokes FEMAG in current workdir and returns pid

        Args:
            options: list of FEMAG options
        """
        if self.cmd.startswith('wfemag') and \
           '-b' in options and \
           '-m' not in options:
            options.insert(0, '-m')

        args = [self.cmd] + options

        if self.__is_running():
            if restart:
                logger.info("must restart")
                self.quit(True)

                # check if process really finished (mq_connection)
                if procId:
                    logger.info("procId: %s", procId)
                    for t in range(200):
                        time.sleep(0.1)
                        if not self.__is_process_running(procId):
                            break
                        logger.info(
                            "femag (pid: '{}') not stopped yet".format(
                                procId))
                    logger.info("Stopped procId: %s", procId)
            else:
                try:
                    with open(os.path.join(self.workdir,
                                           'femag.pid'), 'r') as pidfile:
                        procId = int(pidfile.readline())
                except Exception:
                    pass
                return procId

        basename = str(self.port)
        args.append(basename)

        outname = os.path.join(self.workdir, basename+'.out')
        errname = os.path.join(self.workdir, basename+'.err')
        with open(outname, 'w') as out, open(errname, 'w') as err:
            logger.debug('invoking %s', ' '.join(args))
            self.proc = subprocess.Popen(
                args,
                stdout=out, stderr=err, cwd=self.workdir)
            rc = self.proc.poll()
            if rc:
                raise RuntimeError('Process "{}" exited with {}'.format(
                    ' '.join(args), rc))

        # check if mq is ready for listening
        lcount = 10
        for t in range(lcount):
            time.sleep(0.1)
            if self.__is_running():
                logger.info("femag (pid: '{}') is listening".format(
                    self.proc.pid))
                break
            else:
                # reopen request socket
                logger.debug('reopen request port')
                self.request_socket.close()
                self.request_socket = None
                if t == (lcount-1):
                    # 10 attempts fails, abort
                    logger.info('abort (starting femag)')
                    self.proc.wait(1000)
                    rc = self.proc.returncode
                    self.proc = None
                    return rc  # self.proc.returncode
                self.request_socket = self.__req_socket()

        # write femag.pid
        with open(os.path.join(self.workdir, 'femag.pid'), 'w') as pidfile:
            pidfile.write("{}\n".format(self.proc.pid))
        return self.proc.pid

    def quit(self, save_model=False):
        """terminates femag"""

        if not self.__is_running():
            logger.info("Femag already stopped")
            if self.proc:
                self.proc.wait()
                self.proc = None
            return

        # send exit flags
        f = '\n'.join(['exit_on_end = true',
                       'exit_on_error = true'])
        response = self.send_fsl(f, pub_consumer=subscribe_dev_null)

        # send quit command
        try:
            response = [r.decode('latin1')
                        for r in self.send_request(
                                ['CONTROL', 'quit'], timeout=10000,
                                pub_consumer=subscribe_dev_null)]
        except Exception as e:
            logger.error("Femag Quit zmq message %s", e)

        logger.debug("Sent QUIT to femag %s", response)
        # if query, send a answer
        obj = json.loads(response[0])
        logger.debug("status: {}".format(obj['status']))
        if obj['status'] == 'Query':
            logger.info('query: %s => %s',
                        obj['message'], 'saved' if save_model else 'not saved')

            # Only send one msg
            response = self.request_socket.send_string(
                'Ok' if save_model else 'Cancel')
        
        if self.proc:
            self.proc.wait()
            self.proc = None
        return response

    def upload(self, files):
        """upload file or files
        returns number of transferred bytes
        (FEMAG 8.5 Rev 3282 or greater only)
        """
        ret = []
        total = 0
        chunk_size = 20*1024
        fnames = files
        if isinstance(files, str):
            fnames = [files]
        for fn in fnames:
            self.request_socket.send_string('CONTROL', flags=zmq.SNDMORE)
            basename = os.path.basename(fn)
            logger.info("upload %s --> %s", fn, basename)
            self.request_socket.send_string('upload = {}'.format(basename),
                                            flags=zmq.SNDMORE)
            with open(fn, mode="rb") as file:
                while True:
                    data = file.read(chunk_size)
                    if not data:
                        break
                    more = 0 if len(data) < chunk_size else zmq.SNDMORE
                    self.request_socket.send(data, flags=more)
                    total += len(data)
            ret += [s.decode('latin1')
                    for s in self.request_socket.recv_multipart()]
        return ret
    
    def cleanup(self, timeout=2000):
        """remove all FEMAG files in working directory 
        (FEMAG 8.5 Rev 3282 or greater only)"""
        return [r.decode('latin1')
                for r in self.send_request(['CONTROL', 'cleanup'], timeout=timeout)]
    
    def release(self):
        """signal finish calculation task to load balancer to free resources
        (Docker Cloud environment only)
        """
        return [r.decode('latin1')
                for r in self.send_request(['close'])]

    def info(self, timeout=2000):
        """get various resource information 
        (FEMAG 8.5 Rev 3282 or greater only)"""
        return [r.decode('latin1')
                for r in self.send_request(['CONTROL', 'info'],
                                           timeout=timeout)]

    def publishLevel(self, level):
        """set publish level"""
        return [r.decode('latin1')
                for r in self.send_request(['CONTROL', 'publish = {}'
                                            .format(level)], timeout=10000)]

    def getfile(self, filename=''):
        """get file (FEMAG 8.5 Rev 3282 or greater only)"""
        response = self.send_request(
            ['CONTROL', 'getfile = {}'.format(filename)])
        return [response[0].decode('latin1'),
                response[1] if len(response) else b'']
        
    def exportsvg(self, fslcmds):
        """get svg format from fsl commands (if any graphic created)
        (since FEMAG 8.5 Rev 3343) """
        response = self.send_request(['SVG', fslcmds])
        rc = json.loads(response[0].decode('latin1'))
        if rc['status'] == 'ok':
            return self.getfile(rc['result_file'][0])
        return [s.decode('latin1') for s in response]
            
    def restartStreamReader(self, pub_consumer):
        if self.reader:
            logger.debug("stop stream reader")
            self.reader.continue_loop = False
            self.reader.join()

        if self.reader or (pub_consumer and pub_consumer != subscribe_dev_null):
            self.reader = FemagReadStream(self.__sub_socket(), pub_consumer)
            self.reader.setDaemon(True)
            self.reader.start()

    def copy_magnetizing_curves(self, model, dir=None):
        """extract mc names from model and write files into workdir or dir if given
           and upload to Femag

        Return:
            list of extracted mc names (:obj:`list`)
        """
        dest = dir if dir else self.workdir
        for m in model.set_magcurves(
                self.magnetizingCurves, self.magnets):
            f = self.magnetizingCurves.writefile(m[0], dest, fillfac=m[1])
            self.upload(os.path.join(dest, f))
    
    def __call__(self, pmMachine, simulation, pub_consumer=None):
        """setup fsl file, run calculation and return BCH results
        Args:
          pmMachine: dict with machine parameters or name of model
          simulation; dict with simulation parameters

        Raises:
           FemagError
        """
        if isinstance(pmMachine, str):
            modelpars = dict(name=pmMachine)
        else:
            modelpars = pmMachine
        if 'exit_on_end' not in modelpars:
            modelpars['exit_on_end'] = 'false'
        if 'exit_on_error' not in modelpars:
            modelpars['exit_on_error'] = 'false'
        response = self.send_fsl(['save_model("close")'] +
                                 self.create_fsl(modelpars,
                                                 simulation),
                                 pub_consumer=pub_consumer)
        r = json.loads(response[0])
        if r['status'] != 'ok':
            raise FemagError(r['message'])

        result_file = r['result_file'][0]
        if simulation['calculationMode'] == "pm_sym_loss":
            return self.read_los(self.modelname)

        status, content = self.getfile(result_file)
        r = json.loads(status)
        if r['status'] == 'ok':
            bch = femagtools.bch.Reader()
            bch.read(content.decode('latin1'))
            if simulation['calculationMode'] == 'pm_sym_fast':
                if simulation.get('shortCircuit', False):
                    logger.info("Short Circuit")
                    simulation.update(
                        get_shortCircuit_parameters(bch,
                                                    simulation.get('initial', 2)))
                    builder = femagtools.fsl.Builder()
                    response = self.send_fsl(builder.create_shortcircuit(simulation),
                                             pub_consumer=pub_consumer)
                    r = json.loads(response[0])
                    if r['status'] != 'ok':
                        raise FemagError(r['message'])
                    result_file = r['result_file'][0]
                    status, content = self.getfile(result_file)
                    r = json.loads(status)
                    if r['status'] == 'ok':
                        bch.read(content.decode('latin1'))
                    
            return bch
        raise FemagError(r['message'])


class FemagReadStream(Thread):
    def __init__(self, sub_socket, pub_consumer):
        Thread.__init__(self)
        logger.debug("Initialize reader thread")
        self.sub_socket = sub_socket
        # TODO: use a function to handle calls without pub_consumer
        self.pub_consumer = pub_consumer if pub_consumer else self.dummy_pub_consumer
        self.continue_loop = True

    def dummy_pub_consumer(self, data):
        """This dummy method is used when no pub_consumer is defined.
        """
        logger.warn(
            "No pub_consumer defined, fallback to dummy\n >{}".format(data))

    def run(self):
        """Listen for messages from femag as long as the continue_loop is True
        the sub_socket has a timeout to finish the loop and thread after
        the continue_loop is set to False and no messages are arrived.
        """
        logger.debug(
            "Start thread with while condition: {}".format(self.continue_loop))
        while self.continue_loop:
            try:
                response = self.sub_socket.recv_multipart()
                # Sometimes femag send messages with only len = 1. These messages must be ignored
                if len(response) < 2:
                    continue
                # Call the pub_consumer function
                self.pub_consumer([s.decode('latin1')
                                   for s in response])
            # The subscriber_socket has a timeout of 900 mil sec. If no answer was arrived
            # this exception is raised - Ignore
            except zmq.error.Again:
                continue
            # Any other exception is shown in the error log
            except Exception as e:
                logger.error("error in reading output from femag: {}".format(e))
                continue
        logger.debug("Exit reader thread")



# -*- coding: utf-8 -*-
"""
    femagtools.isa7
    ~~~~~~~~~~~~~~~

    Read FEMAG I7/ISA7 model files
"""
import logging
import struct
import sys
import pdb
import re
import numpy as np
from collections import Counter

logger = logging.getLogger('femagtools.isa7')


class Reader(object):
    """
    Open and Read I7/ISA7 file

    Arguments:
        filename: name of I7/ISA7 file to be read
    """

    def __init__(self, filename):
        with open(filename, mode="rb") as self.file:
            self.file = self.file.read()
        self.pos = 0
        (self.NUM_PNT, self.PNT_PTR, self.PNT_HIDX,
         self.NUM_LIN, self.LIN_PTR, self.LIN_HIDX,
         self.NUM_NOD, self.NOD_PTR, self.NOD_HIDX,
         self.NUM_NDEL, self.NDEL_PTR, self.NDEL_HIDX,
         self.NUM_NDCH, self.NDCH_PTR, self.NDCH_HIDX,
         self.NUM_ELE, self.ELE_PTR, self.ELE_HIDX,
         self.NUM_ELND, self.ELND_PTR, self.ELND_HIDX,
         self.NUM_SPEL, self.SPEL_PTR, self.SPEL_HIDX,
         self.NUM_SE_EL, self.SE_EL_PTR, self.SE_EL_HIDX,
         self.NUM_SPEL_NDCH, self.SPEL_NDCH_PTR, self.SPEL_NDCH_HIDX,
         self.NUM_SR, self.SR_PTR, self.SR_HIDX,
         self.NUM_SR_SE, self.SR_SE_PTR, self.SR_SE_HIDX,
         self.NUM_WB, self.WB_PTR, self.WB_HIDX,
         self.NUM_WB_SR, self.WB_SR_PTR, self.WB_SR_HIDX,
         self.NUM_OB, self.OB_PTR, self.OB_HIDX,
         self.NUM_OB_SR, self.OB_SR_PTR, self.OB_SR_HIDX,
         self.NUM_DV, self.DV_PTR, self.DV_HIDX,
         self.NUM_DV_OB, self.DV_OB_PTR, self.DV_OB_HIDX,
         self.NUM_MC, self.MC_PTR, self.MC_HIDX,
         self.NUM_CF, self.CF_PTR, self.CF_HIDX,
         self.NUM_CF_MC, self.CF_MC_PTR, self.CF_MC_HIDX,
         self.NUM_WN, self.WN_PTR, self.WN_HIDX,
         self.NUM_WN_SW, self.WN_SW_PTR, self.WN_SW_HIDX
         ) = self.next_block("i")

        (valid,
         self.POINT_ISA_POINT_REC_PT_CO_X,
         self.POINT_ISA_POINT_REC_PT_CO_Y) = self.next_block("?ff")

        (valid,
         self.LINE_ISA_LINE_REC_LN_PNT_1,
         self.LINE_ISA_LINE_REC_LN_PNT_2) = self.next_block("?hh")

        (valid,
         self.NODE_ISA_NOD_EL_PNTR,
         self.NODE_ISA_ND_CO_RAD,
         self.NODE_ISA_ND_CO_PHI,
         self.NODE_ISA_NODE_REC_ND_BND_CND,
         self.NODE_ISA_NODE_REC_ND_PER_NOD,
         self.NODE_ISA_NODE_REC_ND_SV_PNTR,
         self.NODE_ISA_NODE_REC_ND_CO_1,
         self.NODE_ISA_NODE_REC_ND_CO_2,
         self.NODE_ISA_NODE_REC_ND_VP_RE,
         self.NODE_ISA_NODE_REC_ND_VP_IM) = self.next_block("?iffhiiffff")

        (self.NOD_ELE_ISA_EL_KEY,
         self.NOD_ELE_ISA_NXT_EL_PNTR) = self.next_block("ii")

        (valid,
         self.NDCHN_ISA_NDCHN_REC_NC_NOD_1,
         self.NDCHN_ISA_NDCHN_REC_NC_NOD_2,
         self.NDCHN_ISA_NDCHN_REC_NC_NOD_MID) = self.next_block("?iii")

        (valid,
         self.ELEM_ISA_EL_NOD_PNTR,
         self.ELEM_ISA_ELEM_REC_EL_TYP,
         self.ELEM_ISA_ELEM_REC_EL_SE_KEY,
         self.ELEM_ISA_ELEM_REC_EL_RELUC,
         self.ELEM_ISA_ELEM_REC_EL_RELUC_2,
         self.ELEM_ISA_ELEM_REC_EL_MAG_1,
         self.ELEM_ISA_ELEM_REC_EL_MAG_2) = self.next_block("?ihhffff")

        (self.ELE_NOD_ISA_ND_KEY,
         self.ELE_NOD_ISA_NXT_ND_PNTR) = self.next_block("ii")

        (valid,
         self.SUPEL_ISA_SE_NDCHN_PNTR,
         self.SUPEL_ISA_SE_EL_PNTR,
         self.SUPEL_ISA_SUPEL_REC_SE_COL,
         self.SUPEL_ISA_SUPEL_REC_SE_MCV_TYP,
         self.SUPEL_ISA_SUPEL_REC_SE_COND_TYP,
         self.SUPEL_ISA_SUPEL_REC_SE_VEL_SYS,
         self.SUPEL_ISA_SUPEL_REC_SE_SR_KEY,
         self.SUPEL_ISA_SUPEL_REC_SE_VELO_1,
         self.SUPEL_ISA_SUPEL_REC_SE_VELO_2,
         self.SUPEL_ISA_SUPEL_REC_SE_CONDUC,
         self.SUPEL_ISA_SUPEL_REC_SE_LENGHT,
         self.SUPEL_ISA_SUPEL_REC_SE_CURD_RE,
         self.SUPEL_ISA_SUPEL_REC_SE_CURD_IM
         ) = self.next_block("?iihhhhhffffff")

        (self.SE_NDCHN_ISA_NC_KEY,
         self.SE_NDCHN_ISA_NXT_NC_PNTR) = self.next_block("ii")

        (self.SE_EL_ISA_EL_KEY,
         self.SE_EL_ISA_NXT_EL_PNTR) = self.next_block("ii")

        (valid,
         self.SR_ISA_SR_SE_PNTR,
         self.SR_ISA_SR_REC_SR_TYP,
         self.SR_ISA_SR_REC_SR_COL,
         self.SR_ISA_SR_REC_SR_NAME,
         self.SR_ISA_SR_REC_SR_CUR_DIR,
         self.SR_ISA_SR_REC_SR_WB_KEY,
         self.SR_ISA_SR_REC_SR_NTURNS,
         self.SR_ISA_SR_REC_SR_SV_PNTR,
         self.SR_ISA_SR_REC_SR_ARRAY,
         self.SR_ISA_SR_REC_SR_GCUR_RE,
         self.SR_ISA_SR_REC_SR_GCUR_IM,
         self.SR_ISA_SR_REC_SR_VOLT_RE,
         self.SR_ISA_SR_REC_SR_VOLT_IM) = self.next_block("?hhh4shhhhfffff")

        (self.SR_SE_ISA_SE_KEY,
         self.SR_SE_ISA_NXT_SE_PNTR) = self.next_block("hh")

        (valid,
         self.WB_ISA_WB_SR_PNTR,
         self.WB_ISA_WB_REC_WB_COL,
         self.WB_ISA_WB_REC_WB_NAME,
         self.WB_TURN,
         self.WB_ISA_WB_REC_WB_SR_NUM,
         self.WB_ISA_WB_REC_WB_WND_KEY,
         self.WB_ISA_WB_REC_WB_UNIT_RES,
         self.WB_ISA_WB_REC_WB_GCUR_RE,
         self.WB_ISA_WB_REC_WB_GCUR_IM,
         self.WB_ISA_WB_REC_WB_VOLT_RE,
         self.WB_ISA_WB_REC_WB_VOLT_IM,
         self.WB_ISA_WB_REC_WB_IMPDZ_RE,
         self.WB_ISA_WB_REC_WB_IMPDZ_IM) = self.next_block("?hh4shhhfffffff")

        self.WB_ISA_WB_REC_WB_TURN = []
        for wd in range(self.NUM_WB):
            if self.WB_ISA_WB_REC_WB_UNIT_RES[wd] == 0:
                self.WB_ISA_WB_REC_WB_TURN.append(
                    self.WB_TURN[wd])
            else:
                self.WB_ISA_WB_REC_WB_TURN.append(
                    self.WB_ISA_WB_REC_WB_UNIT_RES[wd])
                self.WB_ISA_WB_REC_WB_UNIT_RES[wd] = 0

        (self.WB_SR_ISA_SR_KEY,
         self.WB_SR_ISA_NXT_SR_PNTR) = self.next_block("hh")

        self.skip_block(21)

        ANZAHL_TG = self.next_block("i")[1]

        self.skip_block(7)
        self.skip_block(ANZAHL_TG + 1)
        self.skip_block(1)

        self.FC_RADIUS = self.next_block("f")[0]
        self.skip_block(2)
        self.M_POLES = self.next_block("i")[0]
        self.skip_block(6)
        FC_NUM_CUR_ID, FC_NUM_BETA_ID = self.next_block("i")[0:2]
        if FC_NUM_CUR_ID > 16:
            FC_NUM_CUR_ID = 16

        self.skip_block(3)
        self.skip_block(FC_NUM_CUR_ID * 2)
        self.skip_block(1 + 10 * 5 + 3 + 1 * 5 + 14)

        NUM_FE_EVAL_MOVE_STEP = self.next_block("i")[0]
        if NUM_FE_EVAL_MOVE_STEP < 0:
            NUM_FE_EVAL_MOVE_STEP = 0
            
        self.el_fe_induction_1 = [[[]], [[]], [[]]]
        self.el_fe_induction_2 = [[[]], [[]], [[]]]
        self.eddy_cu_vpot = [[[]], [[]], [[]]]
        self.pos_el_fe_induction = []
        
        if NUM_FE_EVAL_MOVE_STEP > 1:
            self.pos_el_fe_induction = self.next_block("f")
            for i in range(NUM_FE_EVAL_MOVE_STEP + 1):
                self.el_fe_induction_1[0][0].append(self.next_block("h"))
                self.el_fe_induction_2[0][0].append(self.next_block("h"))

        FC_NUM_MOVE_CALC_LOAD_PMS, FC_NUM_FLX = self.next_block("i")[0:2]

        if FC_NUM_MOVE_CALC_LOAD_PMS > 1:
            self.skip_block(4)
            self.skip_block(3 * FC_NUM_FLX)
            self.skip_block()

        FC_NUM_MOVE_NOLOAD_PMS = self.next_block("i")[0]

        if FC_NUM_MOVE_NOLOAD_PMS > 1:
            self.skip_block(4)
            self.skip_block(2 * FC_NUM_FLX)
            self.skip_block()

        if NUM_FE_EVAL_MOVE_STEP > 1:
            for i in range(NUM_FE_EVAL_MOVE_STEP + 1):
                self.eddy_cu_vpot[0][0].append(self.next_block("h"))

        self.skip_block(2)  # start_winkel, end_winkel
        self.skip_block(2 * 5)
        self.skip_block(15)
        self.skip_block(3 * 30 * 30)
        self.skip_block(3)
        self.skip_block(30 * 30)
        self.skip_block(4)
        # stator 3
        self.skip_block(4)
        (yoke_diam, inside_diam,
         slot_height,slot_h1,slot_h2,
         slot_width,slot_r1,slot_r2) = self.next_block("f")[:8]
        self.skip_block(3)
        # magnet sector
        magn_rad, yoke_rad, magn_height = self.next_block("f")[:3]
        self.da2 = 2*magn_rad*1e-3
        self.dy2 = 2*yoke_rad*1e-3
        self.da1 = inside_diam
        self.dy1 = yoke_diam
        self.skip_block(3)
        # windings generation
        (tot_num_slot, num_phases, num_layers, 
         self.NUM_WIRES, self.CURRENT,
         coil_span, num_slots) = self.next_block("f")[:7]
        self.TOT_NUM_SLOT = int(tot_num_slot)
        self.NUM_PHASES = int(num_phases)
        self.NUM_LAYERS = int(num_layers)
        self.NUM_SLOTS = int(num_slots)
        self.COIL_SPAN = coil_span
        
        self.skip_block(1)
        (move_action, arm_length, self.SKEW_ANGLE,
         HI, num_move_ar, self.ANGL_I_UP,
         num_par_wdgs, cur_control) = self.next_block("f")[:8]
        self.NUM_PAR_WDGS = int(num_par_wdgs)
        self.ARM_LENGTH = arm_length*1e-3  # unit is m
        self.skip_block(2)      
        self.skip_block(30 * 30)
        self.skip_block(30 * 30)
        self.skip_block(1 * 20)
        self.skip_block(10)
        FC_NUM_MOVE_LOSSES = self.next_block("i")[0]

        if FC_NUM_MOVE_LOSSES > 1 and NUM_FE_EVAL_MOVE_STEP > 1:
            for i in range(NUM_FE_EVAL_MOVE_STEP + 1):
                self.el_fe_induction_1[1][0].append(self.next_block("h"))
                self.el_fe_induction_2[1][0].append(self.next_block("h"))
            for i in range(NUM_FE_EVAL_MOVE_STEP + 1):
                self.eddy_cu_vpot[1][0].append(self.next_block("h"))
            
        # VIRGIN_PM_SYN
        self.skip_block(3)
        # magnet iron 4
        self.skip_block(3)
        # stator 4
        self.skip_block(1)
        # stator 2
        self.skip_block(3)
        # stator 1
        self.skip_block(2)
# ---
        self.skip_block(62)

        ANZ_FORCE_AREAS = self.next_block("i")[0]

        if ANZ_FORCE_AREAS > 3:
            ANZ_FORCE_AREAS = 3

        self.skip_block()
        self.skip_block(2 * ANZ_FORCE_AREAS)
        self.skip_block(14)
        self.skip_block(2 * 3 + 6 * 100 * 3)
        self.skip_block(30)
        self.skip_block(11 * 4)
        self.skip_block()
        self.skip_block(1 * 4)
        # NOM_CURRENT
        # PR_BASIC_LOSS_DATA
        # TOT_MAGNET_AREA
        # MOVE_EXTERN
        # MOVE_ARMATURE
        self.skip_block(5)

        self.POLPAAR_ZAHL, self.NO_POLES_SIM = self.next_block("i")[:2]
        self.SLOT_WIRE_DIAMETER = self.next_block("f")
        self.SLOT_WIRE_NUMBERS = self.next_block("i")
        self.skip_block(20*(3 + 2 * 20)) # BASE_FREQUENCY ..
        self.skip_block(2) # R_TORQUE .. NUM_NOLOAD_EX_CURRENT_STEPS
        (self.R_CURRENT,
         self.R_LOAD_VOLTAGE,
         self.R_NOLOAD_VOLTAGE) = self.next_block("f")
        x = self.next_block("f")
        self.R_COSPHI = x[0]
        self.R_BETA_OPT = x[1:]
        self.skip_block(10) # R_FLUX_LOAD. NUM_NOLOAD_EX_CURRENT_STEPS

        if (FC_NUM_MOVE_LOSSES > 2 and NUM_FE_EVAL_MOVE_STEP > 1
            and FC_NUM_BETA_ID > 1):
            for i in range(NUM_FE_EVAL_MOVE_STEP + 1):
                self.el_fe_induction_1[2][0].append(self.next_block("h"))
                self.el_fe_induction_2[2][0].append(self.next_block("h"))
            for i in range(NUM_FE_EVAL_MOVE_STEP + 1):
                self.eddy_cu_vpot[2][0].append(self.next_block("h"))

        self.skip_block()
        self.skip_block(2 * 3)  # MAX_LOSS_EVAL_STEPS
        self.Q_SLOTS_NUMBER, self.M_PHASE_NUMBER = self.next_block("i")[:2]
        self.N_LAYERS_SLOT, self.N_WIRES_PER_SLOT = self.next_block("i")[:2]
        self.skip_block(1)
        self.skip_block(10 * 100) # num_index_cad
        self.skip_block(1 * 100)
        self.skip_block() # index_cad
        self.skip_block(1 * 4) # heat_tranfer_coeff
        self.skip_block(2 * 2)
        self.skip_block()
        self.skip_block(2 * 4)
        self.skip_block(3)
        self.skip_block(1 * 64) # bnodes_mech
        self.skip_block(6)
            
        self.ELEM_ISA_ELEM_REC_LOSS_DENS = self.next_block("f")
        self.skip_block(3)
        self.skip_block(1 * 64)
        self.ROTOR_CUR_EXIST = self.next_block("?")[0]
        self.skip_block(20)  # mcmax = 20
        self.skip_block(4)
        self.NUM_SE_MAGN_KEYS = self.next_block("i")[0]
        
    def next_block(self, fmt):
        """
        Read binary data and return unpacked values according to format string.

        Arguments:
            fmt: Format string (see python struct module)
        """

        fmt_ = fmt.replace("?", "i")

        blockSize = struct.unpack_from("=i", self.file, self.pos)[0]
        self.pos += 4
        try:
            unpacked = struct.iter_unpack("=" + fmt_,
                                          self.file[self.pos:self.pos
                                                    + blockSize])
            unpacked = [x for x in unpacked]

        except AttributeError:  # python 2 has no iter_unpack
            chunksize = struct.calcsize("=" + fmt_)
            offset = self.pos
            unpacked = []
            for j in range(blockSize // chunksize):
                unpacked.append(struct.unpack_from("=" + fmt_,
                                                   self.file,
                                                   offset))
                offset += chunksize
            logger.info("%s: %d %d", fmt_, blockSize, len(unpacked))

        self.pos += blockSize + 4

        fmt_ = ""
        for s in re.findall(r"[0-9]*.|[0-9]*\?", fmt):
            if len(s) > 1 and s[-1] != "s":
                fmt_ += int(s[:-1]) * s[-1]
            else:
                fmt_ += s
        values = []
        for i, dtype in enumerate(re.findall(r"\?|[0-9]*s?", fmt_)[:-1]):
            if dtype == "?":
                values.append([bool(u[i]) for u in unpacked])
            elif "s" in dtype:
                values.append([u[i].decode('latin-1') for u in unpacked])
            else:
                values.append([u[i] for u in unpacked])

        if len(fmt) == 1:
            return values[0]
        else:
            return values

    def skip_block(self, skips=1):
        """
        Proceed to the next block without reading any data.

        Arguments:
            skips: number of blocks to be skipped
        """
        while skips > 0:
            blockSize = struct.unpack_from("=i", self.file, self.pos)[0]
            self.pos += 4 + blockSize + 4
            skips -= 1


class Isa7(object):
    """
    The ISA7 Femag model

    Arguments:
        filename: name of I7/ISA7 file to be read
    """

    color = {1: [1.0, 0.0, 0.0],
             2: [0.0, 1.0, 0.0],
             3: [1.0, 1.0, 0.0],
             4: [0.0, 0.5019607843137255, 1.0],
             5: [0.9803921568627451, 0.0, 1.0],
             6: [0.0, 1.0, 0.8235294117647058],
             7: [1.0, 1.0, 1.0],
             8: [0.0, 0.0, 0.0],
             9: [0.0, 0.0, 0.5882352941176471],
             10: [0.6666666666666666, 0.0, 0.0],
             11: [0.6666666666666666, 1.0, 0.0],
             12: [1.0, 0.6274509803921569, 0.0],
             13: [0.0, 0.0, 1.0],
             14: [0.6666666666666666, 0.0, 1.0],
             15: [0.0, 0.8235294117647058, 1.0],
             16: [0.8274509803921568, 0.8274509803921568, 0.8274509803921568]}

    def __init__(self, reader):
        self.points = [
            Point(x, y) for x, y in zip(reader.POINT_ISA_POINT_REC_PT_CO_X,
                                        reader.POINT_ISA_POINT_REC_PT_CO_Y)]

        self.lines = []
        for pk1, pk2 in zip(reader.LINE_ISA_LINE_REC_LN_PNT_1,
                            reader.LINE_ISA_LINE_REC_LN_PNT_2):
            point1 = self.points[abs(pk1) - 1]
            point2 = self.points[abs(pk2) - 1]
            self.lines.append(Line(point1, point2))

        self.nodes = [
                Node(n + 1,
                     reader.NODE_ISA_NODE_REC_ND_BND_CND[n],
                     reader.NODE_ISA_NODE_REC_ND_PER_NOD[n],
                     reader.NODE_ISA_ND_CO_RAD[n],
                     reader.NODE_ISA_ND_CO_PHI[n],
                     reader.NODE_ISA_NODE_REC_ND_CO_1[n],
                     reader.NODE_ISA_NODE_REC_ND_CO_2[n],
                     reader.NODE_ISA_NODE_REC_ND_VP_RE[n],
                     reader.NODE_ISA_NODE_REC_ND_VP_IM[n])
            for n in range(len(reader.NODE_ISA_NODE_REC_ND_BND_CND))]

        self.nodechains = []
        for nc in range(len(reader.NDCHN_ISA_NDCHN_REC_NC_NOD_1)):
            nd1 = reader.NDCHN_ISA_NDCHN_REC_NC_NOD_1[nc]
            nd2 = reader.NDCHN_ISA_NDCHN_REC_NC_NOD_2[nc]
            ndm = reader.NDCHN_ISA_NDCHN_REC_NC_NOD_MID[nc]
            try:
                node1 = self.nodes[abs(nd1) - 1]
                nodem = self.nodes[ndm - 1]
                node2 = self.nodes[abs(nd2) - 1]

                if nd1 < 0 or nd2 < 0:
                    nodes = node1, nodem, node2
                elif ndm > 0:
                    nodes = node1, nodem, node2
                else:
                    nodes = node1, None, node2

                self.nodechains.append(
                    NodeChain(nc + 1, nodes))
            except IndexError as ex:
                logger.warning('IndexError in nodes')
                raise  # preserve the stack trace
                    
        self.elements = []
        for e in range(len(reader.ELEM_ISA_EL_NOD_PNTR)):
            ndkeys = []
            ndk = reader.ELEM_ISA_EL_NOD_PNTR[e]

            while ndk > 0:
                ndkeys.append(reader.ELE_NOD_ISA_ND_KEY[ndk - 1])
                ndk = reader.ELE_NOD_ISA_NXT_ND_PNTR[ndk - 1]

            vertices = [self.nodes[k - 1] for k in ndkeys]
            try:
                loss_dens = reader.ELEM_ISA_ELEM_REC_LOSS_DENS[e]
            except (IndexError, AttributeError):
                loss_dens = 0
            self.elements.append(
                Element(e + 1,
                        reader.ELEM_ISA_ELEM_REC_EL_TYP[e],
                        reader.ELEM_ISA_ELEM_REC_EL_SE_KEY[e] - 1,
                        vertices,
                        (reader.ELEM_ISA_ELEM_REC_EL_RELUC[e],
                         reader.ELEM_ISA_ELEM_REC_EL_RELUC_2[e]),
                        (reader.ELEM_ISA_ELEM_REC_EL_MAG_1[e],
                         reader.ELEM_ISA_ELEM_REC_EL_MAG_2[e]),
                        loss_dens * 1e3)   # in W/m³
            )

        self.superelements = []
        for se in range(len(reader.SUPEL_ISA_SE_NDCHN_PNTR)):
            nc_keys = []
            nc_ptr = reader.SUPEL_ISA_SE_NDCHN_PNTR[se]

            while nc_ptr > 0:
                nc_keys.append(reader.SE_NDCHN_ISA_NC_KEY[nc_ptr - 1])
                nc_ptr = reader.SE_NDCHN_ISA_NXT_NC_PNTR[nc_ptr - 1]

            nodechains = []
            for nck in nc_keys:
                if nck > 0:
                    nodechains.append(self.nodechains[abs(nck) - 1])
                else:
                    nodechains.append(self.nodechains[abs(nck) - 1].reverse())

            el_keys = []
            el_ptr = reader.SUPEL_ISA_SE_EL_PNTR[se]

            while el_ptr > 0:
                el_keys.append(reader.SE_EL_ISA_EL_KEY[el_ptr - 1])
                el_ptr = reader.SE_EL_ISA_NXT_EL_PNTR[el_ptr - 1]

            elements = []
            for elk in el_keys:
                elements.append(self.elements[elk - 1])

            self.superelements.append(
                SuperElement(se + 1,
                             reader.SUPEL_ISA_SUPEL_REC_SE_SR_KEY[se] - 1,
                             elements,
                             nodechains,
                             reader.SUPEL_ISA_SUPEL_REC_SE_COL[se],
                             nc_keys,
                             reader.SUPEL_ISA_SUPEL_REC_SE_MCV_TYP[se],
                             reader.SUPEL_ISA_SUPEL_REC_SE_COND_TYP[se],
                             reader.SUPEL_ISA_SUPEL_REC_SE_CONDUC[se],
                             reader.SUPEL_ISA_SUPEL_REC_SE_LENGHT[se],
                             reader.SUPEL_ISA_SUPEL_REC_SE_VEL_SYS[se],
                             reader.SUPEL_ISA_SUPEL_REC_SE_VELO_1[se],
                             reader.SUPEL_ISA_SUPEL_REC_SE_VELO_2[se],
                             reader.SUPEL_ISA_SUPEL_REC_SE_CURD_RE[se],
                             reader.SUPEL_ISA_SUPEL_REC_SE_CURD_IM[se]))

        self.subregions = []
        for sr in range(len(reader.SR_ISA_SR_SE_PNTR)):
            se_keys = []
            se_ptr = reader.SR_ISA_SR_SE_PNTR[sr]

            while se_ptr > 0:
                se_keys.append(reader.SR_SE_ISA_SE_KEY[se_ptr - 1])
                se_ptr = reader.SR_SE_ISA_NXT_SE_PNTR[se_ptr - 1]

            superelements = []
            for sek in se_keys:
                superelements.append(self.superelements[sek - 1])

            nodechains = []
            nc_keys = []
            for se in superelements:
                nc_keys.extend([abs(nc.key) for nc in se.nodechains])
            nc_keys = [nck for nck, count
                       in Counter(nc_keys).items() if count < 2]
            for se in superelements:
                nodechains.extend([nc
                                   for nc in se.nodechains
                                   if abs(nc.key) in nc_keys])

            self.subregions.append(
                SubRegion(sr + 1,
                          reader.SR_ISA_SR_REC_SR_TYP[sr],
                          reader.SR_ISA_SR_REC_SR_COL[sr],
                          reader.SR_ISA_SR_REC_SR_NAME[sr],
                          reader.SR_ISA_SR_REC_SR_CUR_DIR[sr],
                          reader.SR_ISA_SR_REC_SR_WB_KEY[sr] - 1,
                          superelements,
                          nodechains))

        self.windings = []
        for wd in range(len(reader.WB_ISA_WB_SR_PNTR)):
            sr_keys = []
            sr_ptr = reader.WB_ISA_WB_SR_PNTR[wd]

            while sr_ptr > 0:
                sr_keys.append(reader.WB_SR_ISA_SR_KEY[sr_ptr - 1])
                sr_ptr = reader.WB_SR_ISA_NXT_SR_PNTR[sr_ptr - 1]

            subregions = []
            for srk in sr_keys:
                subregions.append(self.subregions[srk - 1])

            self.windings.append(
                Winding(wd + 1,
                        reader.WB_ISA_WB_REC_WB_NAME[wd],
                        subregions,
                        reader.WB_ISA_WB_REC_WB_TURN[wd],
                        reader.WB_ISA_WB_REC_WB_GCUR_RE[wd],
                        reader.WB_ISA_WB_REC_WB_GCUR_IM[wd],
                        reader.WB_ISA_WB_REC_WB_IMPDZ_RE[wd],
                        reader.WB_ISA_WB_REC_WB_IMPDZ_IM[wd],
                        reader.WB_ISA_WB_REC_WB_VOLT_RE[wd],
                        reader.WB_ISA_WB_REC_WB_VOLT_IM[wd]))
        logger.info("Total nodes %d elements %d superelements %d subregions %d",
                    len(self.nodes), len(self.elements),
                    len(self.superelements),
                    len(self.subregions))

        # positions of all elements
        self.element_pos = np.array([e.center
                                     for e in self.elements])
        
        self.FC_RADIUS = reader.FC_RADIUS
        self.POLPAAR_ZAHL = reader.POLPAAR_ZAHL
        self.NO_POLES_SIM = reader.NO_POLES_SIM
        self.ARM_LENGTH = reader.ARM_LENGTH*1e-3  # in m
        self.pos_el_fe_induction = reader.pos_el_fe_induction
        self.el_fe_induction_1 = np.asarray(
            [e for e in reader.el_fe_induction_1 if e[0]]).T/1000
        self.el_fe_induction_2 = np.asarray(
            [e for e in reader.el_fe_induction_2 if e[0]]).T/1000
        self.eddy_cu_vpot = np.asarray(
            [e for e in reader.eddy_cu_vpot if e[0]]).T/1000

    def get_subregion(self, name):
        """return subregion by name"""
        for s in self.subregions:
            if s.name == name:
                return s
        raise ValueError('no such subregion "{}" in this model'.format(name))

    def wdg_elements(self):
        """return elements in winding region"""
        return [el for el in self.elements
                if self.superelement.condtype != 0]

    def get_element(self, x, y):
        """return element at pos x,y"""
        k = np.argmin(np.linalg.norm(self.element_pos - (x, y), axis=1))
        return self.elements[k]

    def get_super_element(self, x, y):
        """return superelement at pos x,y"""
        e = self.get_element(x, y)
        try:
            return [s for s in self.superelements
                    if e.key in [se.key for se in s.elements]][0]
        except IndexError:
            return None
    


class Point(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.xy = x, y


class Line(object):
    def __init__(self, p1, p2):
        self.p1 = p1
        self.p2 = p2


class BaseEntity(object):
    def __init__(self, key):
        self.key = key


class Node(BaseEntity):
    def __init__(self, key, bndcnd, pernod, r, phi, x, y, vpot_re, vpot_im):
        super(self.__class__, self).__init__(key)
        self.bndcnd = bndcnd
        self.pernod = pernod
        self.r = r
        self.phi = phi
        self.x = x
        self.y = y
        self.xy = x, y
        self.vpot = vpot_re, vpot_im

    def on_boundary(self):
        return self.bndcnd != 0 or self.pernod != 0


class NodeChain(BaseEntity):
    def __init__(self, key, nodes):
        super(self.__class__, self).__init__(key)
        self.node1 = nodes[0]
        self.nodemid = nodes[1]
        self.node2 = nodes[2]
        if nodes[1] is None:
            self.nodes = (nodes[0], nodes[2])
        else:
            self.nodes = (nodes[0], nodes[1], nodes[2])

    def reverse(self):
        return NodeChain(self.key * (-1),
                         [self.node2, self.nodemid, self.node1])


class Element(BaseEntity):
    def __init__(self, key, el_type,
                 se_key, vertices, reluc, mag, loss_density):
        super(self.__class__, self).__init__(key)
        self.el_type = el_type
        self.se_key = se_key
        self.vertices = vertices
        self.reluc = reluc
        self.mag = mag
        self.loss_density = loss_density
        if el_type == 1:    # Linear triangle
            self.area = ((vertices[2].x - vertices[1].x) *
                         (vertices[0].y - vertices[1].y) -
                         (vertices[2].y - vertices[1].y) *
                         (vertices[0].x - vertices[1].x))/2
        elif el_type == 2:  # Linear rectangle
            self.area = ((vertices[2].x - vertices[1].x) *
                         (vertices[0].y - vertices[1].y) -
                         (vertices[2].y - vertices[1].y) *
                         (vertices[0].x - vertices[1].x) +
                         (vertices[3].x - vertices[2].x) *
                         (vertices[0].y - vertices[2].y) -
                         (vertices[3].y - vertices[2].y) *
                         (vertices[0].x - vertices[2].x))/2
        elif el_type == 3:  # Square triangle
            self.area = ((vertices[4].x - vertices[2].x) *
                         (vertices[0].y - vertices[2].y) -
                         (vertices[4].y - vertices[1].y) *
                         (vertices[0].x - vertices[2].x))/2
        elif el_type == 4:   # Square rectangle
            self.area = ((vertices[4].x - vertices[2].x) *
                         (vertices[0].y - vertices[2].y) -
                         (vertices[4].y - vertices[2].y) *
                         (vertices[0].x - vertices[2].x) +
                         (vertices[6].x - vertices[4].x) *
                         (vertices[0].y - vertices[4].y) -
                         (vertices[6].y - vertices[4].y) *
                         (vertices[0].x - vertices[4].x))/2
        self.center = np.sum(
            [v.xy for v in vertices], axis=0)/len(vertices)
        
    def induction(self):
        """return induction components of this element"""
        ev = self.vertices
        if self.el_type == 1:
            y31 = ev[2].y - ev[0].y
            y21 = ev[1].y - ev[0].y
            x13 = ev[0].x - ev[2].x
            x21 = ev[1].x - ev[0].x
            a21 = ev[1].vpot[0] - ev[0].vpot[0]
            a31 = ev[2].vpot[0] - ev[0].vpot[0]
            delta = self.superelement.length * (y31 * x21 + y21 * x13)

            return ((x13 * a21 + x21 * a31) / delta,
                    (-y31 * a21 + y21 * a31) / delta)

        elif self.el_type == 2:
            y31 = ev[2].y - ev[0].y
            y21 = ev[1].y - ev[0].y
            x13 = ev[0].x - ev[2].x
            x21 = ev[1].x - ev[0].x
            a21 = ev[1].vpot[0] - ev[0].vpot[0]
            a31 = ev[2].vpot[0] - ev[0].vpot[0]
            delta = self.superelement.length * (y31 * x21 + y21 * x13)
            b1_a = (x13 * a21 + x21 * a31) / delta
            b2_a = (y21 * a31 - y31 * a21) / delta

            y31 = ev[0].y - ev[2].y
            y21 = ev[3].y - ev[2].y
            x13 = ev[2].x - ev[0].x
            x21 = ev[3].x - ev[2].x
            a24 = ev[3].vpot[0] - ev[2].vpot[0]
            a34 = ev[0].vpot[0] - ev[2].vpot[0]
            delta = self.superelement.length * (y31 * x21 + y21 * x13)
            b1_b = (x13 * a24 + x21 * a34) / delta
            b2_b = (y21 * a34 - y31 * a24) / delta

            return ((b1_a + b1_b) / 2,
                    (b2_a + b2_b) / 2)

        return (0, 0)

    def demagnetization(self):
        """return demagnetization of this element"""
        if abs(self.mag[0]) > 1e-5 or abs(self.mag[1]) > 1e-5:
            magn = np.sqrt(self.mag[0]**2 + self.mag[1]**2)
            alfa = np.arctan2(self.mag[1], self.mag[0])
            b1, b2 = self.induction()
            bpol = b1 * np.cos(alfa) + b2 * np.sin(alfa)
            hpol = bpol - magn
            if hpol < 0:
                reluc = abs(self.reluc[0]) / (4*np.pi*1e-7 * 1000)
                return abs(hpol * reluc)
        return 0

    def permeability(self):
        """return permeability of this element"""
        if self.reluc[0] < 1:
            return 1 / self.reluc[0]
        return 1

    def loss_density(self):
        return self.loss_density

    def iron_loss_density(self):
        """return loss_density if element in iron (eg. lamination region)"""
        if self.reluc != (1.0, 1.0) and self.mag == (0.0, 0.0):
            return self.loss_density
        return 0
                
    def mag_loss_density(self):
        """return loss_density if element in magnet region"""
        if np.any(self.mag):
            return self.loss_density
        return 0
    
    def wdg_loss_density(self):
        """return loss_density if element in winding region"""
        if self.superelement.subregion:
            if self.superelement.subregion.winding:
                return self.loss_density
        return 0
        

class SuperElement(BaseEntity):
    def __init__(self, key, sr_key, elements, nodechains, color,
                 nc_keys, mcvtype, condtype, conduc, length,
                 velsys, velo_1, velo_2, curd_re, curd_im):
        super(self.__class__, self).__init__(key)
        self.sr_key = sr_key
        self.subregion = None
        self.elements = elements
        for e in elements:
            e.superelement = self
        self.nodechains = nodechains
        self.color = color
        self.nc_keys = nc_keys
        self.mcvtype = mcvtype
        self.condtype = condtype
        self.conduc = conduc
        self.length = length
        self.velsys = velsys
        self.velo = velo_1, velo_2
        self.curd = curd_re, curd_im


class SubRegion(BaseEntity):
    def __init__(self, key, sr_type, color, name, curdir, wb_key,
                 superelements, nodechains):
        super(self.__class__, self).__init__(key)
        self.sr_type = sr_type
        self.color = color
        self.name = name
        self.curdir = curdir
        self.wb_key = wb_key
        self.winding = None
        self.superelements = superelements
        for se in superelements:
            se.subregion = self
        self.nodechains = nodechains

    def elements(self):
        """return elements of this subregion"""
        e = []
        for s in self.superelements:
            e.append(s.elements)
        return e


class Winding(BaseEntity):
    def __init__(self, key, name, subregions, num_turns, cur_re, cur_im,
                 flux_re, flux_im, volt_re, volt_im):
        super(self.__class__, self).__init__(key)
        self.name = name
        self.subregions = subregions
        for sr in subregions:
            sr.winding = self
        self.num_turns = num_turns
        self.cur = cur_re, cur_im
        self.flux = flux_re, flux_im
        self.volt = volt_re, volt_im

    def elements(self):
        """return elements of this winding"""
        e = []
        for s in self.subregions:
            e.append(s.elements)
        return e


def read(filename):
    """
    Read ISA7 file and return ISA7 object.

    Arguments:
        filename: name of I7/ISA7 file to be read
    """
    import os
    ext = os.path.splitext(filename)[-1]
    if not ext:
        ext = '.I7' if sys.platform == 'win32' else '.ISA7'
        filename += ext
    isa = Isa7(Reader(filename))
    return isa


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    else:
        filename = sys.stdin.readline().strip()

    isa = read(filename)



'''
  mxw2msh Convert Maxwell/msh Files to Gmsh/msh

  Usage:

    mxw2msh <infile> <outfile>

  Author: Jonas Knöpfel

'''
import argparse
import re
from collections import namedtuple

parser = argparse.ArgumentParser()
parser.add_argument("infile")
parser.add_argument("outfile")

args = parser.parse_args()

infile = args.infile
outfile = args.outfile

with open(infile) as f:
    f = iter(f.readlines())
        
def next_line(f, n=1):
    for i in range(n):
        line = next(f)
    line = line.strip()
    line = re.split(" |,", line)
    return line

Node = namedtuple("Node", ["x", "y"])
nodes = []

Element = namedtuple("Element", ["n1", "n2", "n3", "entity"])
elements = []

while True:
    try:
        line = next_line(f)
        if line[0] == "$begin":
            
            if line[1] == "'points'":
                line = next_line(f, 4)
                
                while line[0] != "$end":
                    nodes.append(Node(line[3], line[5]))
                    line = next_line(f)
                    
            if line[1] == "'elements'":
                line = next_line(f)
                
                while line[0] != "$end":
                    elements.append(Element(line[3], line[5], line[7], line[9]))
                    line = next_line(f)
                    
    except StopIteration:
        break

msh = ["$MeshFormat\n2.2 0 8\n$EndMeshFormat"]

msh.append("$Nodes")
msh.append(str(len(nodes)))
for i, n in enumerate(nodes):
    msh.append("{} {} {} 0.0".format(i + 1, n.x, n.y))
msh.append("$EndNodes")

msh.append("$Elements")
msh.append(str(len(elements)))
for i, e in enumerate(elements):
    msh.append("{} 2 2 0 {} {} {} {}".format(i + 1, e.entity, e.n1, e.n2, e.n3))
msh.append("$EndElements")
               
with open(outfile, "w") as outfile:
    outfile.write("\n".join(msh))
